
/* This file is to patch the Cairo source code so it will be possible to compile it */
/* with the OpenWatcom compiler */

/* Author: Doodle <doodle@scenergy.dfmk.hu> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


int cairoint(char *path)
{
  FILE *hFile, *hTarget;
  char achBuffer[1024], achBuffer2[1024];
  int bMutexPatchDone;
  int bEnumPatchDone;
  int bLineReplaced;
  int iPatched;

  sprintf(achBuffer, "%s\\src\\cairoint.h", path);
  hFile = fopen(achBuffer, "rt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: File [%s] could not be opened for patching!\n", achBuffer);
    return 0;
  }

  sprintf(achBuffer, "%s\\src\\$$$_cairoint.h", path);
  hTarget = fopen(achBuffer, "wt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: Tempfile for [%s] could not be opened for patching!\n", achBuffer);
    fclose(hFile);
    return 0;
  }

  bMutexPatchDone = 0;
  bEnumPatchDone = 0;
  iPatched = 0;
  while (!feof(hFile))
  {
    bLineReplaced = 0;
    achBuffer[0] = 0;
    fgets(achBuffer, sizeof(achBuffer), hFile);
    if (achBuffer[0]==0) break;

    if (strstr(achBuffer, "#if defined(__OS2__)") == achBuffer)
      bMutexPatchDone = 1;

    if ((strstr(achBuffer, "#ifndef CAIRO_MUTEX_DECLARE") == achBuffer) &&
	(!bMutexPatchDone))
    {
      // Include the OS/2 mutex defines!
      fprintf(hTarget, "#if defined(__OS2__) && !defined(CAIRO_MUTEX_DECLARE)\n");
      fprintf(hTarget, "# define INCL_DOS\n");
      fprintf(hTarget, "# define INCL_ERRORS\n");
      fprintf(hTarget, "# include <os2def.h>\n");
      fprintf(hTarget, "# include <bsedos.h>\n");
      fprintf(hTarget, "# define CAIRO_MUTEX_DECLARE(name) extern HMTX name\n");
      fprintf(hTarget, "# define CAIRO_MUTEX_DECLARE_GLOBAL(name) extern HMTX name\n");
      fprintf(hTarget, "# define CAIRO_MUTEX_LOCK(name) DosRequestMutexSem(name, SEM_INDEFINITE_WAIT)\n");
      fprintf(hTarget, "# define CAIRO_MUTEX_UNLOCK(name) DosReleaseMutexSem(name)\n");
      fprintf(hTarget, "#endif\n");

      bMutexPatchDone = 1;
      iPatched = 1;
    }

    if (strstr(achBuffer, "typedef enum _cairo_clip_mode cairo_clip_mode_t") == achBuffer)
    {
      // Replace this enum typedef with the correct one from cairo-clip-private.h!
      FILE *hClip;

      sprintf(achBuffer, "%s\\src\\cairo-clip-private.h", path);
      hClip = fopen(achBuffer, "rt");
      if (hClip)
      {
        bEnumPatchDone = 0;
	while ((!bEnumPatchDone) && (!feof(hClip)))
	{
          achBuffer[0] = 0;
	  fgets(achBuffer, sizeof(achBuffer), hClip);
	  if (achBuffer[0]==0) break;

	  if (strstr(achBuffer, "enum _cairo_clip_mode") == achBuffer)
	  {
	    // Found it! Copy this enum into hTarget!
	    fprintf(hTarget, "typedef enum cairo_clip_mode {\n");
	    while ((!bEnumPatchDone) && (!feof(hClip)))
	    {
	      fgets(achBuffer, sizeof(achBuffer), hClip);
	      if (strstr(achBuffer, "};"))
	      {
		bEnumPatchDone = 1;
	      } else
		fputs(achBuffer, hTarget);
	    }
            fprintf(hTarget, "} cairo_clip_mode_t;\n");
	    bEnumPatchDone = 1;
            bLineReplaced = 1;
	  }

	}
	fclose(hClip);
      }
      bEnumPatchDone = 1;
      iPatched = 1;
    }

    if (!bLineReplaced)
      fputs(achBuffer, hTarget);
  }

  fclose(hFile);
  fclose(hTarget);

  if (!iPatched)
  {
    // Remove temp file, we did not patch the original one
    sprintf(achBuffer, "%s\\src\\$$$_cairoint.h", path);
    remove(achBuffer);
  } else
  {
    // Remove the original file, then rename the temp file!
    fprintf(stdout, "INFO: File [%s\\src\\cairoint.h] has been patched!\n", path);
    sprintf(achBuffer, "%s\\src\\cairoint.h", path);
    sprintf(achBuffer2, "%s\\src\\$$$_cairoint.h", path);
    remove(achBuffer);
    rename(achBuffer2, achBuffer);
  }
  return 1;
}

int cairoclipprivate(char *path)
{
  FILE *hFile, *hTarget;
  char achBuffer[1024], achBuffer2[1024];
  int iPatched;

  sprintf(achBuffer, "%s\\src\\cairo-clip-private.h", path);
  hFile = fopen(achBuffer, "rt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: File [%s] could not be opened for patching!\n", achBuffer);
    return 0;
  }

  sprintf(achBuffer, "%s\\src\\$$$_cairo-clip-private.h", path);
  hTarget = fopen(achBuffer, "wt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: Tempfile for [%s] could not be opened for patching!\n", achBuffer);
    fclose(hFile);
    return 0;
  }

  iPatched = 0;
  while (!feof(hFile))
  {
    achBuffer[0] = 0;
    fgets(achBuffer, sizeof(achBuffer), hFile);
    if (achBuffer[0]==0) break;

    if (strstr(achBuffer, "enum _cairo_clip_mode") == achBuffer)
    {
      // Found it! Skip this enum!
      while ((!iPatched) && (!feof(hFile)))
      {
	fgets(achBuffer, sizeof(achBuffer), hFile);
	if (strstr(achBuffer, "};"))
	  iPatched = 1;
      }
    }
    else
      fputs(achBuffer, hTarget);
  }

  fclose(hFile);
  fclose(hTarget);

  if (!iPatched)
  {
    // Remove temp file, we did not patch the original one
    sprintf(achBuffer, "%s\\src\\$$$_cairo-clip-private.h", path);
    remove(achBuffer);
  } else
  {
    // Remove the original file, then rename the temp file!
    fprintf(stdout, "INFO: File [%s\\src\\cairo-clip-private.h] has been patched!\n", path);
    sprintf(achBuffer, "%s\\src\\cairo-clip-private.h", path);
    sprintf(achBuffer2, "%s\\src\\$$$_cairo-clip-private.h", path);
    remove(achBuffer);
    rename(achBuffer2, achBuffer);
  }
  return 1;
}

int cairo_funcprep(char *path, char *file, char *funcbegin)
{
  FILE *hFile, *hTarget;
  char achBuffer[1024], achBuffer2[1024];
  int bLineReplaced;
  int iPatched;
  int bFirstLine;

  sprintf(achBuffer, "%s\\src\\%s", path, file);
  hFile = fopen(achBuffer, "rt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: File [%s] could not be opened for patching!\n", achBuffer);
    return 0;
  }

  sprintf(achBuffer, "%s\\src\\$$$_%s", path, file);
  hTarget = fopen(achBuffer, "wt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: Tempfile for [%s] could not be opened for patching!\n", achBuffer);
    fclose(hFile);
    return 0;
  }

  bFirstLine = 1;
  iPatched = 0;
  bLineReplaced = 0;
  sprintf(achBuffer, "");
  while (!feof(hFile))
  {
    // Copy previous line into buffer2
    strcpy(achBuffer2, achBuffer);

    // Read line
    achBuffer[0] = 0;
    fgets(achBuffer, sizeof(achBuffer), hFile);
    if (achBuffer[0]==0) break;

    if ((strstr(achBuffer, funcbegin) == achBuffer) &&
	(strstr(achBuffer, "(")))
    {
      // This is a cairo function define!
      fprintf(hTarget, "CAIROEXPORT %s", achBuffer2);
      fprintf(hTarget, "CAIROCALL   %s", achBuffer);

      bLineReplaced += 2;
      iPatched = 1;
    }

    if ((strstr(achBuffer, "typedef ") == achBuffer) &&
	(strstr(achBuffer, "(*")) &&
	(strstr(achBuffer, "CAIROCALL") == NULL)
       )
    {
      char *pchType, *pchRemaining;

      // This is a cairo callback typedef!
      pchType = &(achBuffer[8]);
      pchRemaining = strstr(achBuffer, "(*");
      *(pchRemaining-1) = 0;

      fputs(achBuffer2, hTarget);
      fprintf(hTarget, "typedef %s CAIROCALL %s", pchType, pchRemaining);

      bLineReplaced += 2;
      iPatched = 1;
    }

    if ((!bLineReplaced) &&
        (!bFirstLine))
      fputs(achBuffer2, hTarget);

    if (bLineReplaced>0)
      bLineReplaced--;
    bFirstLine = 0;
  }

  if (!bLineReplaced)
    fputs(achBuffer2, hTarget);

  fclose(hFile);
  fclose(hTarget);

  if (!iPatched)
  {
    // Remove temp file, we did not patch the original one
    sprintf(achBuffer, "%s\\src\\$$$_%s", path, file);
    remove(achBuffer);
  } else
  {
    // Remove the original file, then rename the temp file!
    fprintf(stdout, "INFO: File [%s\\src\\%s] has been patched!\n", path, file);
    sprintf(achBuffer, "%s\\src\\%s", path, file);
    sprintf(achBuffer2, "%s\\src\\$$$_%s", path, file);
    remove(achBuffer);
    rename(achBuffer2, achBuffer);
  }
  return 1;
}

int cairooutputstream(char *path)
{
  FILE *hFile, *hTarget;
  char achBuffer[1024], achBuffer2[1024];
  int bLineReplaced;
  int iPatched;

  sprintf(achBuffer, "%s\\src\\cairo-output-stream.c", path);
  hFile = fopen(achBuffer, "rt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: File [%s] could not be opened for patching!\n", achBuffer);
    return 0;
  }

  sprintf(achBuffer, "%s\\src\\$$$_cairo-output-stream.c", path);
  hTarget = fopen(achBuffer, "wt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: Tempfile for [%s] could not be opened for patching!\n", achBuffer);
    fclose(hFile);
    return 0;
  }

  iPatched = 0;
  bLineReplaced = 0;
  sprintf(achBuffer, "");
  while (!feof(hFile))
  {
    // Read line
    achBuffer[0] = 0;
    fgets(achBuffer, sizeof(achBuffer), hFile);
    if (achBuffer[0]==0) break;

    if (strstr(achBuffer, "stdio_write") == achBuffer)
    {
      fprintf(hTarget, "CAIROCALL   %s", achBuffer);

      bLineReplaced += 1;
      iPatched = 1;
    }

    if (!bLineReplaced)
      fputs(achBuffer, hTarget);

    if (bLineReplaced>0)
      bLineReplaced--;
  }

  fclose(hFile);
  fclose(hTarget);

  if (!iPatched)
  {
    // Remove temp file, we did not patch the original one
    sprintf(achBuffer, "%s\\src\\$$$_cairo-output-stream.c", path);
    remove(achBuffer);
  } else
  {
    // Remove the original file, then rename the temp file!
    fprintf(stdout, "INFO: File [%s\\src\\cairo-output-stream.c] has been patched!\n", path);
    sprintf(achBuffer, "%s\\src\\cairo-output-stream.c", path);
    sprintf(achBuffer2, "%s\\src\\$$$_cairo-output-stream.c", path);
    remove(achBuffer);
    rename(achBuffer2, achBuffer);
  }
  return 1;
}

int ftconfig(char *path)
{
  FILE *hFile, *hTarget;
  char achBuffer[1024], achBuffer2[1024];
  int iPatched;

  sprintf(achBuffer, "%s\\include\\freetype\\config\\ftconfig.h", path);
  hFile = fopen(achBuffer, "rt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: File [%s] could not be opened for patching!\n", achBuffer);
    return 0;
  }

  sprintf(achBuffer, "%s\\include\\freetype\\config\\$$$_ftconfig.h", path);
  hTarget = fopen(achBuffer, "wt");
  if (!hFile)
  {
    fprintf(stdout, "INFO: Tempfile for [%s] could not be opened for patching!\n", achBuffer);
    fclose(hFile);
    return 0;
  }

  iPatched = 0;
  while (!feof(hFile))
  {
    achBuffer[0] = 0;
    fgets(achBuffer, sizeof(achBuffer), hFile);
    if (achBuffer[0]==0) break;

    if (strstr(achBuffer, "Watcom doesn't provide 64-bit data types"))
    {
      // Found it! Replace it!
      fprintf(hTarget, "  /* this compiler provides the __int64 type */\n");
      fprintf(hTarget, "#define FT_LONG64\n");
      fprintf(hTarget, "#define FT_INT64 __int64\n");

      iPatched = 1;
    }
    else
      fputs(achBuffer, hTarget);
  }

  fclose(hFile);
  fclose(hTarget);

  if (!iPatched)
  {
    // Remove temp file, we did not patch the original one
    sprintf(achBuffer, "%s\\include\\freetype\\config\\$$$_ftconfig.h", path);
    remove(achBuffer);
  } else
  {
    // Remove the original file, then rename the temp file!
    fprintf(stdout, "INFO: File [%s\\include\\freetype\\config\\ftconfig.h] has been patched!\n", path);
    sprintf(achBuffer, "%s\\include\\freetype\\config\\ftconfig.h", path);
    sprintf(achBuffer2, "%s\\include\\freetype\\config\\$$$_ftconfig.h", path);
    remove(achBuffer);
    rename(achBuffer2, achBuffer);
  }
  return 1;
}


int main(int argc, char *argv[])
{
  if (argc<=2)
  {
    printf("Cairo Source Code Patcher v1.2\n");
    printf("Please give me the base directory of\n");
    printf("Cairo and Freetype2 sources\n");
    printf("as two command line parameters!\n");
    return 1;
  }

  // Patch cairoint.h
  if (!cairoint(argv[1]))
    return 1;

  // Patch cairo-clip-private.h
  if (!cairoclipprivate(argv[1]))
    return 1;

  // Patch cairo.h
  if (!cairo_funcprep(argv[1], "cairo.h", "cairo_"))
    return 1;

  // Patch cairo-ft.h
  if (!cairo_funcprep(argv[1], "cairo-ft.h", "cairo_"))
    return 1;

  // Patch cairo-ps.h
  if (!cairo_funcprep(argv[1], "cairo-ps.h", "cairo_"))
    return 1;

  // Patch cairo-pdf.h
  if (!cairo_funcprep(argv[1], "cairo-pdf.h", "cairo_"))
    return 1;

  // Patch cairo-output-stream.c
  if (!cairooutputstream(argv[1]))
    return 1;

  // Patch ftconfig.h
  if (!ftconfig(argv[2]))
    return 1;


  return 0;
}
